from datetime import datetime

__all__ = ["ENV_VARS", "API_BASE_DEFAULT", "API_BASE_V1"]

"""The v1 API base URL"""
API_BASE_V1 = "https://wavy.fm/api/v1beta"

"""The default API base URL"""
API_BASE_DEFAULT = API_BASE_V1

"""Environment variable names"""
ENV_VARS = {
    "client_id": "WAVYFM_CLIENT_ID",
    "client_secret": "WAVYFM_CLIENT_SECRET",
}


def datetime_from_string(raw_date: str) -> datetime:
    """
    Creates a UTC DateTime object from a RFC3339 formatted string.
    """
    return datetime.strptime(raw_date, '%Y-%m-%dT%H:%M:%S.%f%z')



from typing import Dict, Union


class _WavyProfileEndpoints:
    """
    Endpoints for fetching a user's information.
    """

    def __init__(self, client, user_uri: str):
        self._client = client
        self._uri = user_uri

    def __repr__(self):
        return f"<WavyProfileEndpoints '{self._uri}'>"

    def get_profile(self) -> Dict:
        """
        Retrieve this user's profile.
        """
        return self._client._internal_call("GET", f"/users/{self._uri}", None, {})

    def get_history_stats(self) -> Dict:
        """
        The total amount of listens and artists for this user.
        """
        return self._client._internal_call("GET", f"/users/{self._uri}/history/stats", None, {})

    def get_total_listens(self) -> int:
        """
        The total amount of listens this user.
        """
        stats = self.get_history_stats()
        return stats["total_listens"]

    def get_total_artists(self) -> int:
        """
        The total amount of artists this user.
        """
        stats = self.get_history_stats()
        return stats["total_artists"]

    def get_currently_listening(self) -> Dict:
        """
        What the user is currently listening to.
        The 'item' key will be None if the user is not listening to music.

        Note: This endpoint is cached for about 15 seconds, and is somewhat unstable.
        """
        return self._client._internal_call(
            "GET", f"/users/{self._uri}/history/current", None, {})

    def get_recent_listens(self, size: int = 10) -> Dict:
        """
        Retrieves the user's most recently played tracks.

        Note: the model for this endpoint is heavily subject to change.
        :param size: the amount of listens to retrieve. Valid values are 1 <= size <= 50.
        """
        return self._client._internal_call(
            "GET", f"/users/{self._uri}/history/recent", None, {"size": size})


class _WavyUsersEndpoints:
    """
    Endpoints for fetching users.
    """

    def __init__(self, client):
        self._client = client

    def by_uri(self, uri: str) -> _WavyProfileEndpoints:
        """
        Returns the endpoints for this user by URI.
        This DOES NOT make any HTTP calls.
        """
        return _WavyProfileEndpoints(self._client, uri)

    def by_id(self, user_id: str) -> _WavyProfileEndpoints:
        """
        Returns the endpoints for this user by wavy.fm User ID.
        This DOES NOT make any HTTP calls.
        """
        return _WavyProfileEndpoints(self._client, f"wavyfm:user:id:{user_id}")

    def by_username(self, username: str) -> _WavyProfileEndpoints:
        """
        Returns the endpoints for this user by wavy.fm username.
        This DOES NOT make any HTTP calls.
        """
        return _WavyProfileEndpoints(self._client, f"wavyfm:user:username:{username}")

    def by_discord_id(self, snowflake: Union[str, int]) -> _WavyProfileEndpoints:
        """
        Returns the endpoints for this user by Discord ID (snowflake).
        This DOES NOT make any HTTP calls.
        """
        return _WavyProfileEndpoints(self._client, f"wavyfm:user:discord:{snowflake}")



class WavyException(Exception):
    """An error!"""

    def __init__(self,
                 message: str,
                 error_status: int = None,
                 error_code=None,
                 error_detail=None,
                 *args, **kwargs):
        self.error_name = message
        self.error_code = error_code
        self.error_status = error_status
        self.error_detail = error_detail
        self.__dict__.update(kwargs)
        super().__init__(message, *args)



from typing import List, Dict


class _WavyMetricsEndpoints:
    """
    Global metrics for wavy.fm.
    """

    def __init__(self, client):
        self._client = client

    def get_total_listens(self) -> int:
        """
        The total amount of recorded listens on wavy.fm.
        """
        return self._client._internal_call("GET", "/metrics/total-listens", None, {})

    def get_total_users(self) -> int:
        """
        The total amount of registered users on wavy.fm.
        """
        return self._client._internal_call("GET", "/metrics/total-users", None, {})

    def get_user_listens_leaderboard(self) -> List[Dict]:
        """
        Get the top 10 registered users on wavy.fm by listen count.
        """
        return self._client._internal_call("GET", "/metrics/user-listens-leaderboard", None, {})



"""
Authentication methods for wavy.fm.

This code is heavily inspired from Spotipy: https://github.com/plamere/spotipy
"""

__all__ = [
    "WavyAuthError",
    "WavyClientCredentials",
    "_WavyAuthBase",
]

import base64
import logging
import os
import time
from typing import Optional, Union, Dict

import requests

from wavyfm.error import WavyException
from wavyfm.util import ENV_VARS, API_BASE_DEFAULT

logger = logging.getLogger(__name__)


class WavyAuthError(WavyException):
    """An error during authentication"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class _WavyAuthBase(object):
    def __init__(self, requests_session: Union[requests.Session, bool] = None):
        if isinstance(requests_session, requests.Session):
            self._session = requests_session
        elif requests_session:  # Build a new session if it's a truthy value.
            self._session = requests.Session()
        else:  # Use the Requests API module as a "session".
            from requests import api
            self._session = api

    @property
    def client_id(self) -> Optional[str]:
        return self._client_id

    @client_id.setter
    def client_id(self, client_id: str):
        self._client_id = _ensure_value(client_id, "client_id")

    @property
    def client_secret(self) -> Optional[str]:
        return self._client_secret

    @client_secret.setter
    def client_secret(self, client_secret: str):
        self._client_secret = _ensure_value(client_secret, "client_secret")

    @staticmethod
    def is_token_expired(token_info):
        now = int(time.time())
        # leave 30 seconds of grace
        return token_info["expires_at"] - now < 30

    def get_access_token(self) -> str:
        raise NotImplementedError()

    def __del__(self):
        # Make sure the connection gets closed
        if isinstance(self._session, requests.Session):
            self._session.close()


class WavyClientCredentials(_WavyAuthBase):
    TOKEN_ENDPOINT = f"{API_BASE_DEFAULT}/token"

    def __init__(self,
                 client_id: str = None,
                 client_secret: str = None,
                 proxies: str = None,
                 requests_session: Union[requests.Session, bool] = True,
                 requests_timeout=None):
        """
        Initializes credentials for the Client Credentials flow. You can
        either provide a client_id and client_secret to this constructor,
        or set WAVYFM_CLIENT_ID and WAVYFM_CLIENT_SECRET environment variables.

        To get a good understanding of this flow, you should read the
        wavy.fm Authorization Guide: https://wavy.fm/developers/docs/topics/auth

        :param client_id: Your app's Client ID, obtained from the app dashboard.
        :param client_secret: Your app's Client Secret, obtained from the app dashboard.
        :param proxies: Refer to the `requests` library (:obj:`requests.Session`).
        :param requests_session: An optional :obj:`requests.Session` instance to use.
                                 If set to `True`, a session will be crafted for you.
        :param requests_timeout: Tell Requests to stop waiting for a response after a given number
                                 of seconds.
        """

        super().__init__(requests_session)

        self.client_id = client_id
        self.client_secret = client_secret
        self.token_info = None
        self.proxies = proxies
        self.requests_timeout = requests_timeout

    def get_access_token(self) -> str:
        """
        If a valid access token is in memory, this function returns it.
        If the token is expired, a new token is automatically requested.

        :return: The access token (str)
        """

        if self.token_info and not _WavyAuthBase.is_token_expired(self.token_info):
            return self.token_info["access_token"]

        token_info = self._request_access_token()
        token_info["expires_at"] = int(time.time()) + token_info["expires_in"]
        self.token_info = token_info
        return self.token_info["access_token"]

    def _request_access_token(self) -> Dict:
        """Requests a client credentials access token"""
        payload = {"grant_type": "client_credentials"}
        headers = self._make_authorization_headers()

        logger.debug(
            "sending POST request to %s with Headers: %s and Body: %r",
            self.TOKEN_ENDPOINT, headers, payload
        )

        response = self._session.post(
            self.TOKEN_ENDPOINT,
            data=payload,
            headers=headers,
            verify=True,
            proxies=self.proxies,
            timeout=self.requests_timeout,
        )

        if response.status_code == 200:
            return response.json()
        else:
            try:
                error_body = response.json()
                raise WavyAuthError(
                    error_body["name"],
                    error_status=response.status_code,
                    error_code=error_body.get("code"),
                    error_detail=error_body.get("detail")
                )
            except ValueError:
                raise WavyAuthError(
                    response.text,
                    error_status=response.status_code,
                    error_code=None,
                    error_detail=None,
                )

    def _make_authorization_headers(self) -> Dict[str, str]:
        """The request headers for authorization"""
        auth_header = base64.b64encode(
            f"{self.client_id}:{self.client_secret}".encode("ascii")
        ).decode("ascii")

        return {"Authorization": f"Basic {auth_header}",
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json"}


def _ensure_value(value: str, env_key: str) -> str:
    env_val = ENV_VARS[env_key]
    _val = value or os.getenv(env_val)
    if _val is None:
        msg = "No %s. Pass it or set a %s environment variable." % (
            env_key,
            env_val,
        )
        raise WavyAuthError(msg)
    return _val



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

test_reqs = [
    'mock==2.0.0'
]

doc_reqs = [
]

extra_reqs = {
    'doc': doc_reqs,
    'test': test_reqs
}

setup(
    name='wavyfm',
    version='1.0.2',
    description='Official Python library for wavy.fm',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Aram Peres",
    author_email="aram.peres@wavy.fm",
    url='https://wavy.fm/developers',
    project_urls={
        'Documentation': 'https://github.com/wavy/wavyfm-python',
        'Source': 'https://github.com/wavy/wavyfm-python',
    },
    install_requires=[
        'requests>=2.20.1'
    ],
    tests_require=test_reqs,
    extras_require=extra_reqs,
    license='LICENSE.md',
    packages=['wavyfm'])



import logging

__all__ = ["WavyClient", "WavyException"]

import json
import requests
from requests.packages import urllib3

from wavyfm.error import WavyException
from wavyfm.metrics import _WavyMetricsEndpoints
from wavyfm.users import _WavyUsersEndpoints
from wavyfm.util import API_BASE_DEFAULT
from wavyfm.auth import _WavyAuthBase

logger = logging.getLogger(__name__)


class WavyClient(object):
    default_retry_codes = (429, 500, 502, 503, 504)

    def __init__(self,
                 requests_session=True,
                 auth: str = None,
                 auth_manager: _WavyAuthBase = None,
                 proxies=None,
                 requests_timeout=5,
                 status_forcelist=None,
                 retries=3,
                 status_retries=3,
                 backoff_factor=0.3):
        self._auth = auth
        self.auth_manager = auth_manager
        self.proxies = proxies
        self.requests_timeout = requests_timeout
        self.retries = retries
        self.backoff_factor = backoff_factor
        self.status_forcelist = status_forcelist or self.default_retry_codes
        self.status_retries = status_retries

        if isinstance(requests_session, requests.Session):
            self._session = requests_session
        elif requests_session:  # Build a new session if it's a truthy value.
            self._session = requests.Session()
        else:  # Use the Requests API module as a "session".
            from requests import api
            self._session = api

        self._metrics = _WavyMetricsEndpoints(self)
        self._users = _WavyUsersEndpoints(self)

    @property
    def metrics(self) -> _WavyMetricsEndpoints:
        return self._metrics

    @property
    def users(self) -> _WavyUsersEndpoints:
        return self._users

    def _build_session(self):
        self._session = requests.Session()
        retry = urllib3.Retry(
            total=self.retries,
            connect=None,
            read=False,
            method_whitelist=frozenset(['GET', 'POST', 'PUT', 'DELETE']),
            status=self.status_retries,
            backoff_factor=self.backoff_factor,
            status_forcelist=self.status_forcelist)

        adapter = requests.adapters.HTTPAdapter(max_retries=retry)
        self._session.mount('http://', adapter)
        self._session.mount('https://', adapter)

    def _auth_headers(self):
        if self._auth:
            return {"Authorization": f"Bearer {self._auth}"}
        if not self.auth_manager:
            return {}
        token = self.auth_manager.get_access_token()
        return {"Authorization": f"Bearer {token}"}

    def _internal_call(self, method: str, url: str, payload, params):
        args = dict(params=params)
        if not url.startswith("http"):
            url = API_BASE_DEFAULT + url
        headers = self._auth_headers()

        if "content_type" in args["params"]:
            headers["Content-Type"] = args["params"]["content_type"]
            del args["params"]["content_type"]
            if payload:
                args["data"] = payload
        else:
            headers["Content-Type"] = "application/json"
            if payload:
                args["data"] = json.dumps(payload)

        logger.debug('Sending %s to %s with Params: %s Headers: %s and Body: %r ',
                     method, url, args.get("params"), headers, args.get('data'))

        try:
            response = self._session.request(
                method, url, headers=headers, proxies=self.proxies,
                timeout=self.requests_timeout, **args
            )

            response.raise_for_status()
            results = response.json()
        except requests.exceptions.HTTPError as http_error:
            response = http_error.response
            try:
                msg = response.json()["name"]
            except (ValueError, KeyError):
                msg = "error"
            try:
                code = response.json()["code"]
            except (ValueError, KeyError):
                code = None
            try:
                detail = response.json()["detail"]
            except (ValueError, KeyError):
                detail = None

            logger.error('HTTP Error for %s to %s returned %s due to %s',
                         method, url, response.status_code, msg)

            raise WavyException(
                msg,
                error_status=response.status_code,
                error_code=code,
                error_detail=detail,
            )
        except requests.exceptions.RetryError as retry_error:
            logger.error('Max Retries reached')
            try:
                reason = retry_error.args[0].reason
            except (IndexError, AttributeError):
                reason = None
            raise WavyException(
                reason,
                error_status=-1,
            )
        except ValueError:
            results = None

        return results

    def __del__(self):
        # Make sure the connection gets closed
        if isinstance(self._session, requests.Session):
            self._session.close()
